package fr.she3py.nameprefixes;

import java.util.Locale;

import net.luckperms.api.LuckPerms;
import net.luckperms.api.cacheddata.CachedMetaData;
import net.luckperms.api.event.EventBus;
import net.luckperms.api.event.user.UserDataRecalculateEvent;
import net.luckperms.api.model.user.User;

import org.bukkit.ChatColor;
import org.bukkit.Server;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.player.PlayerEvent;
import org.bukkit.event.player.PlayerJoinEvent;
import org.bukkit.event.player.PlayerTeleportEvent;
import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.scoreboard.Scoreboard;
import org.bukkit.scoreboard.ScoreboardManager;
import org.bukkit.scoreboard.Team;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.Nullable;

public class NamePrefixes extends JavaPlugin implements Listener {
	private String prefix;
	private ChatColor color;
	
	private @Nullable Scoreboard scoreboard;
	private LuckPerms luckPerms;
	
	@SuppressWarnings("resource") // auto closed when plugin is disabled
	@Override
	public void onEnable() {
		this.saveDefaultConfig();
		this.prefix = this.getConfig().getString("format", "§4#ERROR §r");
		
		try {
			this.color = ChatColor.valueOf(this.getConfig().getString("color", ChatColor.RESET.name()));
		}
		catch(IllegalArgumentException ignored) {
			//noinspection StringConcatenation
			this.getLogger().severe("Invalid color in config.yml: " + this.getConfig().getString("color"));
			this.color = ChatColor.RESET;
		}
		
		final Server server = this.getServer();
		
		//noinspection ConstantConditions
		this.luckPerms = server.getServicesManager().load(LuckPerms.class);
		if(this.luckPerms == null) {
			throw new IllegalStateException("LuckPerms unavailable");
		}
		
		final EventBus bus = this.luckPerms.getEventBus();
		bus.subscribe(this, UserDataRecalculateEvent.class, e -> this.getServer().getScheduler().runTask(this, () -> this.onUserUpdate(e)));
		
		server.getPluginManager().registerEvents(this, this);
	}
	
	@NonNls
	private static String makeTeamName(String prefix) {
		return "__NamePrefixes_" + prefix.replace(' ', '_') + '_' + Integer.toHexString(prefix.hashCode());
	}
	
	private static boolean isMe(Team team) {
		return team.getName().startsWith("__NamePrefixes_");
	}
	
	/**
	 * Get or create the team associated with the specified prefix.
	 */
	private Team slurpTeam(String prefix, Scoreboard scoreboard) {
		final String name = makeTeamName(prefix);
		
		Team team = scoreboard.getTeam(name);
		if(team == null) {
			team = scoreboard.registerNewTeam(name);
			team.setPrefix(ChatColor.translateAlternateColorCodes('&', String.format(Locale.ROOT, this.prefix, prefix)));
			team.setColor(this.color);
			team.setAllowFriendlyFire(true);
			team.setCanSeeFriendlyInvisibles(false);
			team.setOption(Team.Option.NAME_TAG_VISIBILITY, Team.OptionStatus.ALWAYS);
			team.setOption(Team.Option.DEATH_MESSAGE_VISIBILITY, Team.OptionStatus.ALWAYS);
			team.setOption(Team.Option.COLLISION_RULE, Team.OptionStatus.ALWAYS);
		}
		
		return team;
	}
	
	private void updatePlayer(Player player, User user, CachedMetaData meta) {
		final ScoreboardManager scoreboardManager = this.getServer().getScoreboardManager();
		if(scoreboardManager == null) {
			return; // no world are loaded
		}
		
		//this.getLogger().info(player.getName() + "'s scoreboard is " + Integer.toHexString(player.getScoreboard().hashCode()) + ", main scoreboard is " + Integer.toHexString(scoreboardManager.getMainScoreboard().hashCode()) + ", plugin scoreboard is " + ((this.scoreboard != null) ? Integer.toHexString(this.scoreboard.hashCode()) : "<null>"));
		
		if(player.getScoreboard() == scoreboardManager.getMainScoreboard()) {
			if(this.scoreboard == null) {
				this.scoreboard = scoreboardManager.getNewScoreboard();
			}
			
			// avoid team "memory leaks" since Bukkit don't allow us to remove teams
			player.setScoreboard(this.scoreboard);
		}
		
		final Scoreboard scoreboard = player.getScoreboard();
		final String entry = player.getName();
		final Team team = scoreboard.getEntryTeam(entry);
		
		final String prefix = meta.getPrefix();
		
		if(prefix == null || prefix.isBlank()) {
			if(team != null) {
				team.removeEntry(entry);
			}
			
			return;
		}
		
		final Team wantedTeam = slurpTeam(prefix, scoreboard);
		if(team != null && team != wantedTeam) {
			if(isMe(team)) {
				team.removeEntry(entry);
				wantedTeam.addEntry(entry);
			}
			else {
				//noinspection StringConcatenation
				this.getLogger().severe("Incompatible team `" + team.getName() + "` with `" + wantedTeam.getName() + "`!");
			}
		}
		else if(team == null) {
			wantedTeam.addEntry(entry);
		}
	}
	
	// LuckPerm event bus
	private void onUserUpdate(UserDataRecalculateEvent e) {
		final User user = e.getUser();
		final Player player = this.getServer().getPlayer(user.getUniqueId());
		
		if(player != null) {
			updatePlayer(player, user, e.getData().getMetaData());
		}
	}
	
	private void handleGenericPlayerEvent(PlayerEvent e) {
		final Player player = e.getPlayer();
		
		var adapter = this.luckPerms.getPlayerAdapter(Player.class);
		updatePlayer(player, adapter.getUser(player), adapter.getMetaData(player));
	}
	
	@EventHandler(priority = EventPriority.HIGHEST)
	public void onPlayerJoin(PlayerJoinEvent e) {
		handleGenericPlayerEvent(e);
	}
	
	@EventHandler(priority = EventPriority.HIGHEST)
	public void onPlayerTeleport(PlayerTeleportEvent e) {
		handleGenericPlayerEvent(e);
	}
}
